/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.topology;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;

import org.ow2.petals.topology.generated.ObjectFactory;
import org.ow2.petals.topology.generated.Topology;
import org.xml.sax.SAXException;

/**
 * The topology builder allows to unmarshall the topology.xml file.
 * @author Nicolas SALATGE - EBM WebSourcing
 */
public final class TopologyBuilder {

    /**
     * XSD schema
     */
    public static final String NAMESPACE_XSD = "http://www.w3.org/2001/XMLSchema";

    /**
     * Topology schema
     */
    public static final String TOPOLOGY_XSD = "petalsTopology.xsd";

    /**
     * The exception raised during JAXB context creation
     */
    private static TopologyException jaxbContextCreationEx;

    /**
     * The unmarshaller.
     */
    private static Unmarshaller unmarshaller;

    /**
     * The exception raised during JAXB unique unmarshaller creation.
     */
    private static TopologyException unmarshCreationEx;

    /**
     * The marshaller.
     */
    private static Marshaller marshaller;
    
    /**
     * The exception raised during JAXB unique marshaller creation.
     */
    private static TopologyException marshCreationEx;

    /**
     * Static object initialization
     */
    static {

        final InputStream schemaInputStream = TopologyBuilder.class.getResourceAsStream("/"
                + TopologyBuilder.TOPOLOGY_XSD);

        try {
            final SchemaFactory factory = SchemaFactory.newInstance(TopologyBuilder.NAMESPACE_XSD);
            final Source[] schemas = new StreamSource[] { new StreamSource(schemaInputStream) };
            final Schema schema = factory.newSchema(schemas);

            try {
                final JAXBContext jaxbContext = JAXBContext.newInstance(new Class[]{Topology.class});

                try {

                    TopologyBuilder.unmarshaller = jaxbContext.createUnmarshaller();
                    TopologyBuilder.unmarshaller.setSchema(schema);

                } catch (final JAXBException ex) {
                    TopologyBuilder.unmarshCreationEx = new TopologyException(
                            "Failed to create the JAXB unmarshaller", ex);
                }

                try {

                    TopologyBuilder.marshaller = jaxbContext.createMarshaller();
                    TopologyBuilder.marshaller.setSchema(schema);
                    TopologyBuilder.marshaller.setProperty(javax.xml.bind.Marshaller.JAXB_ENCODING, "UTF-8");
                    TopologyBuilder.marshaller.setProperty(javax.xml.bind.Marshaller.JAXB_FORMATTED_OUTPUT, java.lang.Boolean.TRUE);

                } catch (final JAXBException ex) {
                    TopologyBuilder.marshCreationEx = new TopologyException(
                            "Failed to create the JAXB marshaller", ex);
                }
            } catch (JAXBException ex) {
                TopologyBuilder.jaxbContextCreationEx = new TopologyException(
                        "Failed to create the JAXB context", ex);
            }

        } catch (final SAXException e) {
            final TopologyException schemaCreationEx = new TopologyException(
                    "Failed to parse resource '" + TopologyBuilder.TOPOLOGY_XSD + "'", e);
            TopologyBuilder.jaxbContextCreationEx = new TopologyException(
                    "Failed to create the JAXB context", schemaCreationEx);
            TopologyBuilder.unmarshCreationEx = new TopologyException(
                    "Failed to create the JAXB unmarshaller", schemaCreationEx);
            TopologyBuilder.marshCreationEx = new TopologyException(
                    "Failed to create the JAXB marshaller", schemaCreationEx);
        }
    }

    /**
     * No constructor available.
     */
    private TopologyBuilder() {
        // NOP
    }

    /**
     * Create the topology object.
     * 
     * @param xmlFileName
     *      the xml file name
     * @return the topology
     * @throws TopologyException
     *      impossible to create the topology
     */
    public static Topology createTopology(final String xmlFileName) throws TopologyException {
        return TopologyBuilder.createTopology(new File(xmlFileName));
    }

    /**
     * Create the topology object.
     * 
     * @param xmlFile
     *      the xml file
     * @return the topology
     * @throws TopologyException
     *      impossible to create the topology
     */
    public static Topology createTopology(final File xmlFile) throws TopologyException {
        try {
            final FileInputStream fis = new FileInputStream(xmlFile);
            try {
                return TopologyBuilder.createTopology(fis);
            } finally {
                try {
                    fis.close();
                } catch (final IOException e) {
                    // NOP
                }
            }
        } catch (final FileNotFoundException e) {
            throw new TopologyException("The file '" + xmlFile.getAbsolutePath()
                    + "' does not exist", e);
        }
    }

    /**
     * Create the topology object.
     * 
     * @param inputStrem
     *            an input stream containing the XML instance of the topology.
     *            The input stream <b>MUST</b> be closed by the caller.
     * @return the topology
     * @throws TopologyException
     *             impossible to create the topology
     */
    public static Topology createTopology(final InputStream inputStream) throws TopologyException {

        if (TopologyBuilder.unmarshaller == null) {
            throw TopologyBuilder.unmarshCreationEx;
        }
        
        try {
            final JAXBElement<Topology> root;
            // The default Xerces unmarshaller is not thread safe
            synchronized (TopologyBuilder.unmarshaller) {
                root = TopologyBuilder.unmarshaller.unmarshal(new StreamSource(inputStream),
                        Topology.class);
            }

            return root.getValue();
        } catch (final JAXBException e) {
            throw new TopologyException("Failed to build topology", e);
        }
    }

    /**
     * Write the topology file to the disk
     * 
     * @param xmlFile
     * @param topology
     */
    public synchronized static void marshallTopologyFile(final File xmlFile, final Topology topology) throws TopologyException {

        if (TopologyBuilder.marshaller == null) {
            throw TopologyBuilder.marshCreationEx;
        }
        
        try {
            ObjectFactory jaxbObjectFactory = new ObjectFactory();

            JAXBElement<Topology> topologyElement = jaxbObjectFactory.createTopology(topology);

            TopologyBuilder.marshaller.marshal(topologyElement, xmlFile);
        } catch (JAXBException ex) {
            throw new TopologyException("Failed to write local topology on disk", ex);
        }
        
    }
}
